# Provide a nicer error message than failing to import models.Index.

VERSION = (0, 5, 2)
__version__ = '.'.join(str(v) for v in VERSION)


__all__ = ['PartialIndex', 'PQ', 'PF', 'ValidatePartialUniqueMixin', 'PartialUniqueValidationError']


MIN_DJANGO_VERSION = (1, 11)
DJANGO_VERSION_ERROR = 'Django version %s or later is required for django-partial-index.' % '.'.join(str(v) for v in MIN_DJANGO_VERSION)

try:
    import django
except ImportError:
    raise ImportError(DJANGO_VERSION_ERROR)

if tuple(django.VERSION[:2]) < MIN_DJANGO_VERSION:
    raise ImportError(DJANGO_VERSION_ERROR)


from .index import PartialIndex
from .query import PQ, PF
from .mixins import ValidatePartialUniqueMixin, PartialUniqueValidationError



from django.db.models import Index, Q
from django.utils import six
import warnings

from . import query


def validate_where(where='', where_postgresql='', where_sqlite=''):
    if where:
        if where_postgresql or where_sqlite:
            raise ValueError('If providing a common where predicate, must not provide where_postgresql or where_sqlite.')
        if isinstance(where, six.string_types):
            warnings.warn(
                'Text-based where predicates are deprecated, will be removed in a future release. ' +
                'Please upgrade to where=PQ().',
                DeprecationWarning)
        elif isinstance(where, query.PQ):
            pass
        else:
            raise ValueError('Where predicate must be a string or a partial_index.PQ object.')
    else:
        if not where_postgresql and not where_sqlite:
            raise ValueError('A where predicate must be provided.')
        if where_postgresql == where_sqlite:
            raise ValueError('If providing a separate where_postgresql and where_sqlite, then they must be different.' +
                             'If the same expression works for both, just use single where.')
        if not isinstance(where_postgresql, six.string_types) or not isinstance(where_sqlite, six.string_types):
            raise ValueError('where_postgresql and where_sqlite must be strings.')
        warnings.warn(
            'Text-based where predicates are deprecated, will be removed in a future release. ' +
            'Please upgrade to where=PQ().',
            DeprecationWarning)
    return where, where_postgresql, where_sqlite


class PartialIndex(Index):
    suffix = 'partial'
    # Allow an index name longer than 30 characters since this index can only be used on PostgreSQL and SQLite,
    # and the Django default 30 character limit for cross-database compatibility isn't applicable.
    # The "partial" suffix is 4 letters longer than the default "idx".
    max_name_length = 34
    sql_create_index = {
        'postgresql': 'CREATE%(unique)s INDEX %(name)s ON %(table)s%(using)s (%(columns)s)%(extra)s WHERE %(where)s',
        'sqlite': 'CREATE%(unique)s INDEX %(name)s ON %(table)s%(using)s (%(columns)s) WHERE %(where)s',
    }

    # Mutable default fields=[] looks wrong, but it's copied from super class.
    def __init__(self, fields=[], name=None, unique=None, where='', where_postgresql='', where_sqlite=''):
        if unique not in [True, False]:
            raise ValueError('Unique must be True or False')
        self.unique = unique
        self.where, self.where_postgresql, self.where_sqlite = \
            validate_where(where=where, where_postgresql=where_postgresql, where_sqlite=where_sqlite)
        super(PartialIndex, self).__init__(fields=fields, name=name)

    def __repr__(self):
        if self.where:
            if isinstance(self.where, query.PQ):
                anywhere = "where=%s" % repr(self.where)
            else:
                anywhere = "where='%s'" % self.where
        else:
            anywhere = "where_postgresql='%s', where_sqlite='%s'" % (self.where_postgresql, self.where_sqlite)

        return "<%(name)s: fields=%(fields)s, unique=%(unique)s, %(anywhere)s>" % {
            'name': self.__class__.__name__,
            'fields': "'{}'".format(', '.join(self.fields)),
            'unique': self.unique,
            'anywhere': anywhere
        }

    def deconstruct(self):
        path, args, kwargs = super(PartialIndex, self).deconstruct()
        if path.startswith('partial_index.index'):
            path = path.replace('partial_index.index', 'partial_index')
        kwargs['unique'] = self.unique
        if self.where:
            kwargs['where'] = self.where
        else:
            kwargs['where_postgresql'] = self.where_postgresql
            kwargs['where_sqlite'] = self.where_sqlite
        return path, args, kwargs

    def get_sql_create_template_values(self, model, schema_editor, using):
        # This method exists on Django 1.11 Index class, but has been moved to the SchemaEditor on Django 2.0.
        # This makes it complex to call superclass methods and avoid duplicating code.
        # Can be simplified if Django 1.11 support is dropped one day.

        # Copied from Django 1.11 Index.get_sql_create_template_values(), which does not exist in Django 2.0:
        fields = [model._meta.get_field(field_name) for field_name, order in self.fields_orders]
        tablespace_sql = schema_editor._get_index_tablespace_sql(model, fields)
        quote_name = schema_editor.quote_name
        columns = [
            ('%s %s' % (quote_name(field.column), order)).strip()
            for field, (field_name, order) in zip(fields, self.fields_orders)
        ]
        parameters = {
            'table': quote_name(model._meta.db_table),
            'name': quote_name(self.name),
            'columns': ', '.join(columns),
            'using': using,
            'extra': tablespace_sql,
        }

        # PartialIndex updates:
        parameters['unique'] = ' UNIQUE' if self.unique else ''
        # Note: the WHERE predicate is not yet checked for syntax or field names, and is inserted into the CREATE INDEX query unescaped.
        # This is bad for usability, but is not a security risk, as the string cannot come from user input.
        vendor = query.get_valid_vendor(schema_editor)
        if isinstance(self.where, query.PQ):
            parameters['where'] = query.q_to_sql(self.where, model, schema_editor)
        elif vendor == 'postgresql':
            parameters['where'] = self.where_postgresql or self.where
        elif vendor == 'sqlite':
            parameters['where'] = self.where_sqlite or self.where
        else:
            raise ValueError('Should never happen')
        return parameters

    def create_sql(self, model, schema_editor, using=''):
        vendor = query.get_valid_vendor(schema_editor)
        sql_template = self.sql_create_index[vendor]
        sql_parameters = self.get_sql_create_template_values(model, schema_editor, using)
        return sql_template % sql_parameters

    def name_hash_extra_data(self):
        return [str(self.unique), self.where, self.where_postgresql, self.where_sqlite]

    def set_name_with_model(self, model):
        """Sets an unique generated name for the index.

        PartialIndex would like to only override "hash_data = ...", but the entire method must be duplicated for that.
        """
        table_name = model._meta.db_table
        column_names = [model._meta.get_field(field_name).column for field_name, order in self.fields_orders]
        column_names_with_order = [
            (('-%s' if order else '%s') % column_name)
            for column_name, (field_name, order) in zip(column_names, self.fields_orders)
        ]
        # The length of the parts of the name is based on the default max
        # length of 30 characters.
        hash_data = [table_name] + column_names_with_order + [self.suffix] + self.name_hash_extra_data()
        self.name = '%s_%s_%s' % (
            table_name[:11],
            column_names[0][:7],
            '%s_%s' % (self._hash_generator(*hash_data), self.suffix),
        )
        assert len(self.name) <= self.max_name_length, (
            'Index too long for multiple database support. Is self.suffix '
            'longer than 3 characters?'
        )
        self.check_name()



from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db.models import Q

from .index import PartialIndex
from . import query


class PartialUniqueValidationError(ValidationError):
    pass


class ValidatePartialUniqueMixin(object):
    """PartialIndex with unique=True validation to ModelForms and Django Rest Framework Serializers.

    Mixin should be added before the parent model class, for example:

    class MyModel(ValidatePartialUniqueMixin, models.Model):
        ...

        indexes = [
            PartialIndex(...)
        ]

    The mixin is usable only for PartialIndexes with a Q-object where-condition. If applied to a model
    with a text-based where-condition, an error is raised.

    Important Note:
    Django's standard ModelForm validation for unique constraints is sub-optimal. If a field belonging to the
    unique index is not present on the form, then it the constraint is not validated. This requires adding
    hidden fields on the form, checking them against tampering, etc.

    ValidatePartialUniqueMixin does not follow that example:
    It always validates with all fields, even if they are not on the form.
    """

    def validate_unique(self, exclude=None):
        # Standard unique validation first.
        super(ValidatePartialUniqueMixin, self).validate_unique(exclude=exclude)
        self.validate_partial_unique()

    def validate_partial_unique(self):
        """Check partial unique constraints on the model and raise ValidationError if any failed.

        We want to check if another instance already exists with the fields mentioned in idx.fields, but only if idx.where matches.
        But can't just check for the fields in idx.fields - idx.where may refer to other fields on the current (or other) models.
        Also can't check for all fields on the current model - should not include irrelevant fields which may hide duplicates.

        To find potential conflicts, we need to build a queryset which:
        1. Filters by idx.fields with their current values on this instance,
        2. Filters on idx.where
        3. Filters by fields mentioned in idx.where, with their current values on this instance,
        4. Excludes current object if it does not match the where condition.

        Note that step 2 ensures the lookup only looks for conflicts among rows covered by the PartialIndes,
        and steps 2+3 ensures that the QuerySet is empty if the PartialIndex does not cover the current object.
        """
        # Find PartialIndexes with unique=True defined on model.
        unique_idxs = [idx for idx in self._meta.indexes if isinstance(idx, PartialIndex) and idx.unique]

        if unique_idxs:
            model_fields = set(f.name for f in self._meta.get_fields(include_parents=True, include_hidden=True))

            for idx in unique_idxs:
                where = idx.where
                if not isinstance(where, Q):
                    raise ImproperlyConfigured(
                        'ValidatePartialUniqueMixin is not supported for PartialIndexes with a text-based where condition. ' +
                        'Please upgrade to Q-object based where conditions.'
                    )

                mentioned_fields = set(idx.fields) | set(query.q_mentioned_fields(where, self.__class__))

                missing_fields = mentioned_fields - model_fields
                if missing_fields:
                    raise RuntimeError('Unable to use ValidatePartialUniqueMixin: expecting to find fields %s on model. ' +
                                       'This is a bug in the PartialIndex definition or the django-partial-index library itself.')

                values = {field_name: getattr(self, field_name) for field_name in mentioned_fields}

                conflict = self.__class__.objects.filter(**values)  # Step 1 and 3
                conflict = conflict.filter(where)  # Step 2
                if self.pk:
                    conflict = conflict.exclude(pk=self.pk)  # Step 4

                if conflict.exists():
                    raise PartialUniqueValidationError('%s with the same values for %s already exists.' % (
                        self.__class__.__name__,
                        ', '.join(sorted(idx.fields)),
                    ))



#!/usr/bin/env python

from setuptools import setup


setup(
    name='django-partial-index',
    packages=['partial_index'],
    version='0.5.2',
    description='PostgreSQL and SQLite partial indexes for Django models',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Mattias Linnap',
    author_email='mattias@linnap.com',
    url='https://github.com/mattiaslinnap/django-partial-index',
    download_url='https://github.com/mattiaslinnap/django-partial-index/archive/0.5.2.tar.gz',
    license='BSD',
    install_requires=[],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Database',
        'Topic :: Internet :: WWW/HTTP',
    ]
)



"""Django Q object to SQL string conversion."""
from django.db.models import expressions, Q, F
from django.db.models.sql import Query


class Vendor(object):
    POSTGRESQL = 'postgresql'
    SQLITE = 'sqlite'


class PQ(Q):
    """Compatibility class for Q-objects.

    Django 2.0 Q-objects are suitable on their own, but Django 1.11 needs a better deep equality comparison and a deconstruct() method.

    PartialIndex definitions in model classes should use PQ to avoid problems when upgrading projects.
    """

    def __eq__(self, other):
        """Copied from Django 2.0 django.utils.tree.Node.__eq__()"""
        if self.__class__ != other.__class__:
            return False
        if (self.connector, self.negated) == (other.connector, other.negated):
            return self.children == other.children
        return False

    def deconstruct(self):
        """Copied from Django 2.0 django.db.models.query_utils.Q.deconstruct()"""
        path = '%s.%s' % (self.__class__.__module__, self.__class__.__name__)
        # Keep imports clean in migrations
        if path.startswith('partial_index.query.'):
            path = path.replace('partial_index.query.', 'partial_index.')

        args, kwargs = (), {}
        if len(self.children) == 1 and not isinstance(self.children[0], Q):
            child = self.children[0]
            kwargs = {child[0]: child[1]}
        else:
            args = tuple(self.children)
            if self.connector != self.default:
                kwargs = {'_connector': self.connector}
        if self.negated:
            kwargs['_negated'] = True
        return path, args, kwargs


class PF(F):
    """Compatibility class for F-expressions.

    Django 2.0 F-expressions are suitable on their own, but Django 1.11 a deconstruct() method.

    PartialIndex definitions in model classes should use PF to avoid problems when upgrading projects.
    """

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return self.name == other.name

    def deconstruct(self):
        path = '%s.%s' % (self.__class__.__module__, self.__class__.__name__)
        # Keep imports clean in migrations
        if path.startswith('partial_index.query.'):
            path = path.replace('partial_index.query.', 'partial_index.')

        args = (self.name, )
        kwargs = {}
        return path, args, kwargs



def get_valid_vendor(schema_editor):
    vendor = schema_editor.connection.vendor
    if vendor not in [Vendor.POSTGRESQL, Vendor.SQLITE]:
        raise ValueError('Database vendor %s is not supported by django-partial-index.' % vendor)
    return vendor


def q_to_sql(q, model, schema_editor):
    # Q -> SQL conversion based on code from Ian Foote's Check Constraints pull request:
    # https://github.com/django/django/pull/7615/

    query = Query(model)
    where = query._add_q(q, used_aliases=set(), allow_joins=False)[0]
    connection = schema_editor.connection
    compiler = connection.ops.compiler('SQLCompiler')(query, connection, 'default')
    sql, params = where.as_sql(compiler, connection)
    params = tuple(map(schema_editor.quote_value, params))
    where_sql = sql % params
    return where_sql


def expression_mentioned_fields(exp):
    if isinstance(exp, expressions.Col):
        field = exp.output_field or exp.field  # TODO: which one makes sense to use here?
        if field and field.name:
            return [field.name]
    elif hasattr(exp, 'get_source_expressions'):
        child_fields = []
        for source in exp.get_source_expressions():
            child_fields.extend(expression_mentioned_fields(source))
        return child_fields
    else:
        raise NotImplementedError('Unexpected expression class %s=%s when looking up mentioned fields.' % (exp.__class__.__name__, exp))


def q_mentioned_fields(q, model):
    """Returns list of field names mentioned in Q object.

    Q(a__isnull=True, b=F('c')) -> ['a', 'b', 'c']
    """
    query = Query(model)
    where = query._add_q(q, used_aliases=set(), allow_joins=False)[0]
    return list(sorted(set(expression_mentioned_fields(where))))
